@@ -1,4 +1,3 @@
-# -*- coding: iso-8859-1 -*-
 # Copyright (C) 2000-2014 Bastian Kleineidam
 #
 # This program is free software; you can redistribute it and/or modify
@@ -18,44 +17,41 @@
 Store metadata and options.
 """
 
+from functools import lru_cache
 import os
 import re
-try:  # Python 3
-    from urllib import parse
-    from urllib import request
-except ImportError:  # Python 2
-    import urlparse as parse
-    import urllib as request
+import urllib.parse
+import urllib.request
 import shutil
 import socket
 import _LinkChecker_configdata as configdata
-from .. import (log, LOG_CHECK, get_install_data, fileutil)
+from .. import log, LOG_CHECK, get_install_data, fileutil
 from . import confparse
-from ..decorators import memoized
 from xdg.BaseDirectory import xdg_config_home, xdg_data_home
 
 Version = configdata.version
 ReleaseDate = configdata.release_date
 AppName = configdata.name
-App = AppName+u" "+Version
+App = AppName + " " + Version
 Author = configdata.author
-HtmlAuthor = Author.replace(u' ', u'&nbsp;')
-Copyright = u"Copyright (C) 2000-2014 "+Author
-HtmlCopyright = u"Copyright &copy; 2000-2014 "+HtmlAuthor
-AppInfo = App+u"              "+Copyright
-HtmlAppInfo = App+u", "+HtmlCopyright
+HtmlAuthor = Author.replace(' ', '&nbsp;')
+Copyright = "Copyright (C) 2000-2016 Bastian Kleineidam, 2010-2020 " + Author
+HtmlCopyright = ("Copyright &copy; 2000-2016 Bastian&nbsp;Kleineidam, 2010-2020 "
+                 + HtmlAuthor)
+HtmlAppInfo = App + ", " + HtmlCopyright
 Url = configdata.url
-SupportUrl = u"https://github.com/wummel/linkchecker/issues"
-DonateUrl = u"http://wummel.github.io/linkchecker/donations.html"
-Email = configdata.author_email
-UserAgent = u"Mozilla/5.0 (compatible; %s/%s; +%s)" % (AppName, Version, Url)
-Freeware = AppName+u""" comes with ABSOLUTELY NO WARRANTY!
-This is free software, and you are welcome to redistribute it
-under certain conditions. Look at the file `LICENSE' within this
-distribution."""
+SupportUrl = "https://github.com/linkchecker/linkchecker/issues"
+UserAgent = "Mozilla/5.0 (compatible; %s/%s; +%s)" % (AppName, Version, Url)
+Freeware = (
+    AppName
+    + """ comes with ABSOLUTELY NO WARRANTY!
+This is free software, and you are welcome to redistribute it under
+certain conditions. Look at the file `LICENSE' within this distribution."""
+)
 Portable = configdata.portable
 
-def normpath (path):
+
+def normpath(path):
     """Norm given system path with all available norm or expand functions
     in os.path."""
     expanded = os.path.expanduser(os.path.expandvars(path))
@@ -64,18 +60,22 @@ def normpath (path):
 
 # List Python modules in the form (module, name, version attribute)
 Modules = (
-# required modules
+    # required modules
+    ("bs4", "Beautiful Soup", "__version__"),
+    ("dns.version", "dnspython", "version"),
     ("requests", "Requests", "__version__"),
-# optional modules
-    ("argcomplete", u"Argcomplete", None),
-    ("GeoIP", u"GeoIP", 'lib_version'),   # on Unix systems
-    ("pygeoip", u"GeoIP", 'lib_version'), # on Windows systems
-    ("sqlite3", u"Pysqlite", 'version'),
-    ("sqlite3", u"Sqlite", 'sqlite_version'),
-    ("gconf", u"Gconf", '__version__'),
-    ("meliae", u"Meliae", '__version__'),
+    ("xdg", "PyXDG", "__version__"),
+    # optional modules
+    ("argcomplete", "Argcomplete", None),
+    ("GeoIP", "GeoIP", 'lib_version'),  # on Unix systems
+    ("pygeoip", "GeoIP", 'lib_version'),  # on Windows systems
+    ("sqlite3", "Pysqlite", 'version'),
+    ("sqlite3", "Sqlite", 'sqlite_version'),
+    ("gi", "PyGObject", '__version__'),
+    ("meliae", "Meliae", '__version__'),
 )
 
+
 def get_modules_info():
     """Return unicode string with detected module info."""
     module_infos = []
@@ -90,20 +90,21 @@ def get_modules_info():
             # ignore attribute errors in case library developers
             # change the version information attribute
             module_infos.append(name)
-    return u"Modules: %s" % (u", ".join(module_infos))
+    return "Modules: %s" % (", ".join(module_infos))
 
 
-def get_share_dir ():
+def get_share_dir():
     """Return absolute path of LinkChecker example configuration."""
     return os.path.join(get_install_data(), "share", "linkchecker")
 
 
-def get_share_file (filename, devel_dir=None):
+def get_share_file(filename, devel_dir=None):
     """Return a filename in the share directory.
+
     @param devel_dir: directory to search when developing
-    @ptype devel_dir: string
+    @type devel_dir: string
     @param filename: filename to search for
-    @ptype filename: string
+    @type filename: string
     @return: the found filename or None
     @rtype: string
     @raises: ValueError if not found
@@ -136,12 +137,14 @@ def get_system_cert_file():
 
 def get_certifi_file():
     """Get the SSL certifications installed by the certifi package.
+
     @return: the filename to the cert file
     @rtype: string
     @raises: ImportError when certifi is not installed or ValueError when
              the file is not found
     """
     import certifi
+
     filename = certifi.where()
     if os.path.isfile(filename):
         return filename
@@ -150,63 +153,64 @@ def get_certifi_file():
 
 
 # dynamic options
-class Configuration (dict):
+class Configuration(dict):
     """
     Storage for configuration options. Options can both be given from
     the command line as well as from configuration files.
     """
 
-    def __init__ (self):
+    def __init__(self):
         """
         Initialize the default options.
         """
-        super(Configuration, self).__init__()
-        ## checking options
+        super().__init__()
+        # checking options
         self["allowedschemes"] = []
         self['cookiefile'] = None
         self['robotstxt'] = True
         self["debugmemory"] = False
         self["localwebroot"] = None
-        self["maxfilesizeparse"] = 1*1024*1024
-        self["maxfilesizedownload"] = 5*1024*1024
+        self["maxfilesizeparse"] = 1 * 1024 * 1024
+        self["maxfilesizedownload"] = 5 * 1024 * 1024
         self["maxnumurls"] = None
         self["maxrunseconds"] = None
         self["maxrequestspersecond"] = 10
         self["maxhttpredirects"] = 10
         self["nntpserver"] = os.environ.get("NNTP_SERVER", None)
-        self["proxy"] = request.getproxies()
+        self["proxy"] = urllib.request.getproxies()
         self["sslverify"] = True
         self["threads"] = 10
         self["timeout"] = 60
         self["aborttimeout"] = 300
         self["recursionlevel"] = -1
         self["useragent"] = UserAgent
-        ## authentication
+        # authentication
         self["authentication"] = []
         self["loginurl"] = None
         self["loginuserfield"] = "login"
         self["loginpasswordfield"] = "password"
         self["loginextrafields"] = {}
-        ## filtering
+        # filtering
         self["externlinks"] = []
         self["ignorewarnings"] = []
         self["internlinks"] = []
         self["checkextern"] = False
-        ## plugins
+        # plugins
         self["pluginfolders"] = get_plugin_folders()
         self["enabledplugins"] = []
-        ## output
+        # output
         self['trace'] = False
         self['quiet'] = False
         self["verbose"] = False
         self["warnings"] = True
         self["fileoutput"] = []
         self['output'] = 'text'
-        self["status"] = False
+        self["status"] = True
         self["status_wait_seconds"] = 5
         self['logger'] = None
         self.loggers = {}
         from ..logger import LoggerClasses
+
         for c in LoggerClasses:
             key = c.LoggerName
             self[key] = {}
@@ -216,18 +220,18 @@ class Configuration (dict):
         """Set the status logger."""
         self.status_logger = status_logger
 
-    def logger_new (self, loggername, **kwargs):
+    def logger_new(self, loggername, **kwargs):
         """Instantiate new logger and return it."""
         args = self[loggername]
         args.update(kwargs)
         return self.loggers[loggername](**args)
 
-    def logger_add (self, loggerclass):
+    def logger_add(self, loggerclass):
         """Add a new logger type to the known loggers."""
         self.loggers[loggerclass.LoggerName] = loggerclass
         self[loggerclass.LoggerName] = {}
 
-    def read (self, files=None):
+    def read(self, files=None):
         """
         Read settings from given config files.
 
@@ -253,20 +257,17 @@ class Configuration (dict):
         log.debug(LOG_CHECK, "reading configuration from %s", filtered_cfiles)
         confparse.LCConfigParser(self).read(filtered_cfiles)
 
-    def add_auth (self, user=None, password=None, pattern=None):
+    def add_auth(self, user=None, password=None, pattern=None):
         """Add given authentication data."""
         if not user or not pattern:
-            log.warn(LOG_CHECK,
-            _("missing user or URL pattern in authentication data."))
+            log.warn(
+                LOG_CHECK, _("missing user or URL pattern in authentication data.")
+            )
             return
-        entry = dict(
-            user=user,
-            password=password,
-            pattern=re.compile(pattern),
-        )
+        entry = dict(user=user, password=password, pattern=re.compile(pattern))
         self["authentication"].append(entry)
 
-    def get_user_password (self, url):
+    def get_user_password(self, url):
         """Get tuple (user, password) from configured authentication
         that matches the given URL.
         Both user and password can be None if not specified, or no
@@ -281,7 +282,7 @@ class Configuration (dict):
         """Get dict with limit per connection type."""
         return {key: self['maxconnections%s' % key] for key in ('http', 'https', 'ftp')}
 
-    def sanitize (self):
+    def sanitize(self):
         "Make sure the configuration is consistent."
         if self['logger'] is None:
             self.sanitize_logger()
@@ -293,52 +294,45 @@ class Configuration (dict):
         # set default socket timeout
         socket.setdefaulttimeout(self['timeout'])
 
-    def sanitize_logger (self):
+    def sanitize_logger(self):
         """Make logger configuration consistent."""
         if not self['output']:
             log.warn(LOG_CHECK, _("activating text logger output."))
             self['output'] = 'text'
         self['logger'] = self.logger_new(self['output'])
 
-    def sanitize_loginurl (self):
+    def sanitize_loginurl(self):
         """Make login configuration consistent."""
         url = self["loginurl"]
         disable = False
-        if not self["loginpasswordfield"]:
-            log.warn(LOG_CHECK,
-            _("no CGI password fieldname given for login URL."))
-            disable = True
-        if not self["loginuserfield"]:
-            log.warn(LOG_CHECK,
-            _("no CGI user fieldname given for login URL."))
-            disable = True
         if self.get_user_password(url) == (None, None):
-            log.warn(LOG_CHECK,
-            _("no user/password authentication data found for login URL."))
+            log.warn(
+                LOG_CHECK,
+                _("no user/password authentication data found for login URL."),
+            )
             disable = True
         if not url.lower().startswith(("http:", "https:")):
             log.warn(LOG_CHECK, _("login URL is not a HTTP URL."))
             disable = True
-        urlparts = parse.urlsplit(url)
+        urlparts = urllib.parse.urlsplit(url)
         if not urlparts[0] or not urlparts[1] or not urlparts[2]:
             log.warn(LOG_CHECK, _("login URL is incomplete."))
             disable = True
         if disable:
-            log.warn(LOG_CHECK,
-              _("disabling login URL %(url)s.") % {"url": url})
+            log.warn(LOG_CHECK, _("disabling login URL %(url)s.") % {"url": url})
             self["loginurl"] = None
 
-    def sanitize_proxies (self):
+    def sanitize_proxies(self):
         """Try to read additional proxy settings which urllib does not
         support."""
         if os.name != 'posix':
             return
         if "http" not in self["proxy"]:
-            http_proxy = get_gconf_http_proxy() or get_kde_http_proxy()
+            http_proxy = get_gnome_proxy() or get_kde_http_proxy()
             if http_proxy:
                 self["proxy"]["http"] = http_proxy
         if "ftp" not in self["proxy"]:
-            ftp_proxy = get_gconf_ftp_proxy() or get_kde_ftp_proxy()
+            ftp_proxy = get_gnome_proxy(protocol="FTP") or get_kde_ftp_proxy()
             if ftp_proxy:
                 self["proxy"]["ftp"] = ftp_proxy
 
@@ -372,10 +366,14 @@ def get_user_data():
     @rtype string
     """
     homedotdir = normpath("~/.linkchecker/")
-    userdata = homedotdir if os.path.isdir(homedotdir) \
+    userdata = (
+        homedotdir
+        if os.path.isdir(homedotdir)
         else os.path.join(xdg_data_home, "linkchecker")
+    )
     return userdata
 
+
 def get_plugin_folders():
     """Get linkchecker plugin folders. Default is
     "$XDG_DATA_HOME/linkchecker/plugins/". "~/.linkchecker/plugins/" is also
@@ -403,7 +401,7 @@ def make_userdir(child):
             # Windows forbids filenames with leading dot unless
             # a trailing dot is added.
             userdir += "."
-        os.makedirs(userdir, 0700)
+        os.makedirs(userdir, 0o700)
 
 
 def get_user_config():
@@ -419,88 +417,77 @@ def get_user_config():
     initialconf = normpath(os.path.join(get_share_dir(), "linkcheckerrc"))
     # per user config settings
     homedotfile = normpath("~/.linkchecker/linkcheckerrc")
-    userconf = homedotfile if os.path.isfile(homedotfile) \
+    userconf = (
+        homedotfile
+        if os.path.isfile(homedotfile)
         else os.path.join(xdg_config_home, "linkchecker", "linkcheckerrc")
-    if os.path.isfile(initialconf) and not os.path.exists(userconf) and \
-       not Portable:
+    )
+    if os.path.isfile(initialconf) and not os.path.exists(userconf) and not Portable:
         # copy the initial configuration to the user configuration
         try:
             make_userdir(userconf)
             shutil.copy(initialconf, userconf)
         except Exception as errmsg:
-            msg = _("could not copy initial configuration file %(src)r to %(dst)r: %(errmsg)r")
+            msg = _(
+                "could not copy initial configuration file %(src)r"
+                " to %(dst)r: %(errmsg)r"
+            )
             args = dict(src=initialconf, dst=userconf, errmsg=errmsg)
             log.warn(LOG_CHECK, msg % args)
     return userconf
 
 
-def get_gconf_http_proxy ():
-    """Return host:port for GConf HTTP proxy if found, else None."""
+def get_gnome_proxy(protocol="HTTP"):
+    """Return host:port for a GNOME proxy if found, else None."""
     try:
-        import gconf
+        import gi
+        gi.require_version('Gio', '2.0')
+        from gi.repository import Gio
     except ImportError:
         return None
     try:
-        client = gconf.client_get_default()
-        if client.get_bool("/system/http_proxy/use_http_proxy"):
-            host = client.get_string("/system/http_proxy/host")
-            port = client.get_int("/system/http_proxy/port")
-            if host:
-                if not port:
-                    port = 8080
-                return "%s:%d" % (host, port)
-    except Exception as msg:
-        log.debug(LOG_CHECK, "error getting HTTP proxy from gconf: %s", msg)
-        pass
-    return None
+        schema_id = "org.gnome.system.proxy.%s" % protocol.lower()
+        # If the schema is not installed Gio.Settings.new() causes Trace/breakpoint trap
+        source = Gio.SettingsSchemaSource.get_default()
+        if source is None:
+            log.debug(LOG_CHECK, "No GSettings schemas are installed")
+            return None
+        schema = source.lookup(schema_id, False)
+        if schema is None:
+            log.debug(LOG_CHECK, "%s not installed" % schema_id)
+            return None
 
-
-def get_gconf_ftp_proxy ():
-    """Return host:port for GConf FTP proxy if found, else None."""
-    try:
-        import gconf
-    except ImportError:
-        return None
-    try:
-        client = gconf.client_get_default()
-        host = client.get_string("/system/proxy/ftp_host")
-        port = client.get_int("/system/proxy/ftp_port")
+        settings = Gio.Settings.new(schema_id)
+        if protocol == "HTTP" and not settings.get_boolean("enabled"):
+            return None
+        host = settings.get_string("host")
+        port = settings.get_int("port")
         if host:
             if not port:
                 port = 8080
             return "%s:%d" % (host, port)
     except Exception as msg:
-        log.debug(LOG_CHECK, "error getting FTP proxy from gconf: %s", msg)
-        pass
+        log.debug(LOG_CHECK, "error getting %s proxy from GNOME: %s", (protocol, msg))
     return None
 
 
-def get_kde_http_proxy ():
+def get_kde_http_proxy():
     """Return host:port for KDE HTTP proxy if found, else None."""
-    config_dir = get_kde_config_dir()
-    if not config_dir:
-        # could not find any KDE configuration directory
-        return
     try:
-        data = read_kioslaverc(config_dir)
+        data = read_kioslaverc()
         return data.get("http_proxy")
     except Exception as msg:
         log.debug(LOG_CHECK, "error getting HTTP proxy from KDE: %s", msg)
-        pass
 
 
-def get_kde_ftp_proxy ():
+def get_kde_ftp_proxy():
     """Return host:port for KDE HTTP proxy if found, else None."""
-    config_dir = get_kde_config_dir()
-    if not config_dir:
-        # could not find any KDE configuration directory
-        return
     try:
-        data = read_kioslaverc(config_dir)
+        data = read_kioslaverc()
         return data.get("ftp_proxy")
     except Exception as msg:
         log.debug(LOG_CHECK, "error getting FTP proxy from KDE: %s", msg)
-        pass
+
 
 # The following KDE functions are largely ported and ajusted from
 # Google Chromium:
@@ -533,59 +520,44 @@ def get_kde_ftp_proxy ():
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-def get_kde_config_dir ():
+
+def get_kde_config_dir():
     """Return KDE configuration directory or None if not found."""
-    kde_home = get_kde_home_dir()
-    if not kde_home:
-        # could not determine the KDE home directory
-        return
-    return kde_home_to_config(kde_home)
-
-
-def kde_home_to_config (kde_home):
-    """Add subdirectories for config path to KDE home directory."""
-    return os.path.join(kde_home, "share", "config")
-
-
-def get_kde_home_dir ():
-    """Return KDE home directory or None if not found."""
     if os.environ.get("KDEHOME"):
-        kde_home = os.path.abspath(os.environ["KDEHOME"])
+        home = os.environ.get("KDEHOME")
     else:
         home = os.environ.get("HOME")
-        if not home:
-            # $HOME is not set
+    if not home:
+        log.debug(LOG_CHECK, "KDEHOME and HOME not set")
+        return
+    kde_config_dir = os.path.join(home, ".config")
+    if not os.path.exists(kde_config_dir):
+        kde_config_dir = os.path.join(home, ".kde4", "share", "config")
+        if not os.path.exists(kde_config_dir):
+            log.debug(LOG_CHECK, "%s does not exist" % kde_config_dir)
             return
-        kde3_home = os.path.join(home, ".kde")
-        kde4_home = os.path.join(home, ".kde4")
-        if fileutil.find_executable("kde4-config"):
-            # kde4
-            kde3_file = kde_home_to_config(kde3_home)
-            kde4_file = kde_home_to_config(kde4_home)
-            if os.path.exists(kde4_file) and os.path.exists(kde3_file):
-                if fileutil.get_mtime(kde4_file) >= fileutil.get_mtime(kde3_file):
-                    kde_home = kde4_home
-                else:
-                    kde_home = kde3_home
-            else:
-                kde_home = kde4_home
-        else:
-            # kde3
-            kde_home = kde3_home
-    return kde_home if os.path.exists(kde_home) else None
+    return kde_config_dir
 
 
 loc_ro = re.compile(r"\[.*\]$")
 
-@memoized
-def read_kioslaverc (kde_config_dir):
+
+@lru_cache(1)
+def read_kioslaverc():
     """Read kioslaverc into data dictionary."""
     data = {}
+    kde_config_dir = get_kde_config_dir()
+    if not kde_config_dir:
+        return data
+    in_proxy_settings = False
     filename = os.path.join(kde_config_dir, "kioslaverc")
+    if not os.path.exists(filename):
+        log.debug(LOG_CHECK, "%s does not exist" % filename)
+        return data
     with open(filename) as fd:
         # First read all lines into dictionary since they can occur
         # in any order.
-        for line in  fd:
+        for line in fd:
             line = line.rstrip()
             if line.startswith('['):
                 in_proxy_settings = line.startswith("[Proxy Settings]")
@@ -606,14 +578,14 @@ def read_kioslaverc (kde_config_dir):
     return data
 
 
-def add_kde_proxy (key, value, data):
+def add_kde_proxy(key, value, data):
     """Add a proxy value to data dictionary after sanity checks."""
     if not value or value[:3] == "//:":
         return
     data[key] = value
 
 
-def add_kde_setting (key, value, data):
+def add_kde_setting(key, value, data):
     """Add a KDE proxy setting value to data dictionary."""
     if key == "ProxyType":
         mode = None
@@ -639,7 +611,13 @@ def add_kde_setting (key, value, data):
     elif key == "ftpProxy":
         add_kde_proxy("ftp_proxy", value, data)
     elif key == "ReversedException":
-        data["reversed_bypass"] = bool(value == "true" or int(value))
+        if value == "true":
+            value = True
+        elif value == "false":
+            value = False
+        else:
+            value = int(value) != 0
+        data["reversed_bypass"] = value
     elif key == "NoProxyFor":
         data["ignore_hosts"] = split_hosts(value)
     elif key == "AuthMode":
@@ -647,12 +625,12 @@ def add_kde_setting (key, value, data):
         # XXX todo
 
 
-def split_hosts (value):
+def split_hosts(value):
     """Split comma-separated host list."""
     return [host for host in value.split(", ") if host]
 
 
-def resolve_indirect (data, key, splithosts=False):
+def resolve_indirect(data, key, splithosts=False):
     """Replace name of environment variable with its value."""
     value = data[key]
     env_value = os.environ.get(value)
@@ -665,7 +643,7 @@ def resolve_indirect (data, key, splithosts=False):
         del data[key]
 
 
-def resolve_kde_settings (data):
+def resolve_kde_settings(data):
     """Write final proxy configuration values in data dictionary."""
     if "mode" not in data:
         return
